/*
 * SpanDSP - a series of DSP components for telephony
 *
 * gsm0610_decode.c - GSM 06.10 full rate speech codec.
 *
 * Written by Steve Underwood <steveu@coppice.org>
 *
 * Copyright (C) 2006 Steve Underwood
 *
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * This code is based on the widely used GSM 06.10 code available from
 * http://kbs.cs.tu-berlin.de/~jutta/toast.html
 *
 * $Id: gsm0610_decode.c,v 1.4 2006/06/13 13:29:46 steveu Exp $
 */

/*! \file */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <assert.h>
#include <inttypes.h>
#include <tgmath.h>
#include <stdlib.h>
#include <stdio.h>

#include "spandsp/telephony.h"
#include "spandsp/dc_restore.h"
#include "spandsp/gsm0610.h"

#include "gsm0610_local.h"

/* 4.3 FIXED POINT IMPLEMENTATION OF THE RPE-LTP DECODER */

static void postprocessing(gsm0610_state_t *s, int16_t amp[])
{
    int k;
    int16_t msr;
    int16_t tmp;

    msr = s->msr;
    for (k = 0;  k < GSM0610_FRAME_LEN;  k++)
    {
        tmp = gsm_mult_r(msr, 28180);
        /* De-emphasis */
        msr = gsm_add(amp[k], tmp);
        /* Truncation & upscaling */
        amp[k] = (int16_t) (gsm_add(msr, msr) & 0xFFF8);
    }
    /*endfor*/
    s->msr = msr;
}
/*- End of function --------------------------------------------------------*/

static void decode_a_frame(gsm0610_state_t *s,
                            int16_t LARcr[8],
                            int16_t Ncr[4],
                            int16_t bcr[4],
                            int16_t Mcr[4],
                            int16_t xmaxcr[4],
                            int16_t *xMcr,     // [0..13*4]
                            int16_t amp[GSM0610_FRAME_LEN])
{
    int j;
    int k;
    int16_t erp[40];
    int16_t wt[GSM0610_FRAME_LEN];
    int16_t *drp;
    int16_t *Ncrx;
    int16_t *bcrx;
    int16_t *Mcrx;
    int16_t *xmaxcrx;

    Ncrx = Ncr;
    bcrx = bcr;
    Mcrx = Mcr;
    xmaxcrx = xmaxcr;
    
    drp = s->dp0 + 120;
    for (j = 0;  j < 4;  j++, xMcr += 13)
    {
        gsm0610_rpe_decoding(s, *xmaxcrx++, *Mcrx++, xMcr, erp);
        gsm0610_long_term_synthesis_filtering(s, *Ncrx++, *bcrx++, erp, drp);
        for (k = 0;  k < 40;  k++)
            wt[j*40 + k] = drp[k];
        /*endfor*/
    }
    /*endfor*/

    gsm0610_short_term_synthesis_filter(s, LARcr, wt, amp);
    postprocessing(s, amp);
}
/*- End of function --------------------------------------------------------*/

const uint8_t *gsm0610_decode(gsm0610_state_t *s, int16_t amp[], const uint8_t c[], int quant)
{
    int16_t LARc[8];
    int16_t Nc[4];
    int16_t Mc[4];
    int16_t bc[4];
    int16_t xmaxc[4];
    int16_t xmc[13*4];
    uint16_t sr;
    int i;

    if (s->packing == GSM0610_PACKING_WAV49)
        quant <<= 1;
    while (quant--)
    {
        switch (s->packing)
        {
        case GSM0610_PACKING_NONE:
            /* No packing at all */
            LARc[0] = c[0];
            LARc[1] = c[1];
            LARc[2] = c[2];
            LARc[3] = c[3];
            LARc[4] = c[4];
            LARc[5] = c[5];
            LARc[6] = c[6];
            LARc[7] = c[7];
            Nc[0] = c[8];
            bc[0] = c[9];
            Mc[0] = c[10];
            xmaxc[0] = c[11];
            xmc[0] = c[12];
            xmc[1] = c[13];
            xmc[2] = c[14];
            xmc[3] = c[15];
            xmc[4] = c[16];
            xmc[5] = c[17];
            xmc[6] = c[18];
            xmc[7] = c[19];
            xmc[8] = c[20];
            xmc[9] = c[21];
            xmc[10] = c[22];
            xmc[11] = c[23];
            xmc[12] = c[24];
            Nc[1] = c[25];
            bc[1] = c[26];
            Mc[1] = c[27];
            xmaxc[1] = c[28];
            xmc[13] = c[29];
            xmc[14] = c[30];
            xmc[15] = c[31];
            xmc[16] = c[32];
            xmc[17] = c[33];
            xmc[18] = c[34];
            xmc[19] = c[35];
            xmc[20] = c[36];
            xmc[21] = c[37];
            xmc[22] = c[38];
            xmc[23] = c[39];
            xmc[24] = c[40];
            xmc[25] = c[41];
            Nc[2] = c[42];
            bc[2] = c[43];
            Mc[2] = c[44];
            xmaxc[2] = c[45];
            xmc[26] = c[46];
            xmc[27] = c[47];
            xmc[28] = c[48];
            xmc[29] = c[49];
            xmc[30] = c[50];
            xmc[31] = c[51];
            xmc[32] = c[52];
            xmc[33] = c[53];
            xmc[34] = c[54];
            xmc[35] = c[55];
            xmc[36] = c[56];
            xmc[37] = c[57];
            xmc[38] = c[58];
            Nc[3] = c[59];
            bc[3] = c[60];
            Mc[3] = c[61];
            xmaxc[3] = c[62];
            xmc[39] = c[63];
            xmc[40] = c[64];
            xmc[41] = c[65];
            xmc[42] = c[66];
            xmc[43] = c[67];
            xmc[44] = c[68];
            xmc[45] = c[69];
            xmc[46] = c[70];
            xmc[47] = c[71];
            xmc[48] = c[72];
            xmc[49] = c[73];
            xmc[50] = c[74];
            xmc[51] = c[75];
            c += 76;
            break;
        case GSM0610_PACKING_WAV49:
            s->frame_index = !s->frame_index;
            if (s->frame_index)
            {
                sr = *c++;
                LARc[0] = sr & 0x3F;
                sr >>= 6;
                sr |= (uint16_t) *c++ << 2;
                LARc[1] = sr & 0x3F;
                sr >>= 6;
                sr |= (uint16_t) *c++ << 4;
                LARc[2] = sr & 0x1F;
                sr >>= 5;
                LARc[3] = sr & 0x1F;
                sr >>= 5;
                sr |= (uint16_t) *c++ << 2;
                LARc[4] = sr & 0xF;
                sr >>= 4;
                LARc[5] = sr & 0xF;
                sr >>= 4;
                sr |= (uint16_t) *c++ << 2;                                /* 5 */
                LARc[6] = sr & 0x7;
                sr >>= 3;
                LARc[7] = sr & 0x7;
                sr >>= 3;
                sr |= (uint16_t) *c++ << 4;
                Nc[0] = sr & 0x7f;
                sr >>= 7;
                bc[0] = sr & 0x3;
                sr >>= 2;
                Mc[0] = sr & 0x3;
                sr >>= 2;
                sr |= (uint16_t) *c++ << 1;
                xmaxc[0] = sr & 0x3F;
                sr >>= 6;
                xmc[0] = sr & 0x7;
                sr >>= 3;
                sr = *c++;
                xmc[1] = sr & 0x7;
                sr >>= 3;
                xmc[2] = sr & 0x7;
                sr >>= 3;
                sr |= (uint16_t) *c++ << 2;
                xmc[3] = sr & 0x7;
                sr >>= 3;
                xmc[4] = sr & 0x7;
                sr >>= 3;
                xmc[5] = sr & 0x7;
                sr >>= 3;
                sr |= (uint16_t) *c++ << 1;                                /* 10 */
                xmc[6] = sr & 0x7;
                sr >>= 3;
                xmc[7] = sr & 0x7;
                sr >>= 3;
                xmc[8] = sr & 0x7;
                sr >>= 3;
                sr = *c++;
                xmc[9] = sr & 0x7;
                sr >>= 3;
                xmc[10] = sr & 0x7;
                sr >>= 3;
                sr |= (uint16_t) *c++ << 2;
                xmc[11] = sr & 0x7;
                sr >>= 3;
                xmc[12] = sr & 0x7;
                sr >>= 3;
                sr |= (uint16_t) *c++ << 4;
                Nc[1] = sr & 0x7F;
                sr >>= 7;
                bc[1] = sr & 0x3;
                sr >>= 2;
                Mc[1] = sr & 0x3;
                sr >>= 2;
                sr |= (uint16_t) *c++ << 1;
                xmaxc[1] = sr & 0x3F;
                sr >>= 6;
                xmc[13] = sr & 0x7;
                sr >>= 3;
                sr = *c++;                                                 /* 15 */
                xmc[14] = sr & 0x7;
                sr >>= 3;
                xmc[15] = sr & 0x7;
                sr >>= 3;
                sr |= (uint16_t) *c++ << 2;
                xmc[16] = sr & 0x7;
                sr >>= 3;
                xmc[17] = sr & 0x7;
                sr >>= 3;
                xmc[18] = sr & 0x7;
                sr >>= 3;
                sr |= (uint16_t) *c++ << 1;
                xmc[19] = sr & 0x7;
                sr >>= 3;
                xmc[20] = sr & 0x7;
                sr >>= 3;
                xmc[21] = sr & 0x7;
                sr >>= 3;
                sr = *c++;
                xmc[22] = sr & 0x7;
                sr >>= 3;
                xmc[23] = sr & 0x7;
                sr >>= 3;
                sr |= (uint16_t) *c++ << 2;
                xmc[24] = sr & 0x7;
                sr >>= 3;
                xmc[25] = sr & 0x7;
                sr >>= 3;
                sr |= (uint16_t) *c++ << 4;                               /* 20 */
                Nc[2] = sr & 0x7F;
                sr >>= 7;
                bc[2] = sr & 0x3;
                sr >>= 2;
                Mc[2] = sr & 0x3;
                sr >>= 2;
                sr |= (uint16_t) *c++ << 1;
                xmaxc[2] = sr & 0x3F;
                sr >>= 6;
                xmc[26] = sr & 0x7;
                sr >>= 3;
                sr = *c++;
                xmc[27] = sr & 0x7;
                sr >>= 3;
                xmc[28] = sr & 0x7;
                sr >>= 3;
                sr |= (uint16_t) *c++ << 2;
                xmc[29] = sr & 0x7;
                sr >>= 3;
                xmc[30] = sr & 0x7;
                sr >>= 3;
                xmc[31] = sr & 0x7;
                sr >>= 3;
                sr |= (uint16_t) *c++ << 1;
                xmc[32] = sr & 0x7;
                sr >>= 3;
                xmc[33] = sr & 0x7;
                sr >>= 3;
                xmc[34] = sr & 0x7;
                sr >>= 3;
                sr = *c++;                                                /* 25 */
                xmc[35] = sr & 0x7;
                sr >>= 3;
                xmc[36] = sr & 0x7;
                sr >>= 3;
                sr |= (uint16_t) *c++ << 2;
                xmc[37] = sr & 0x7;
                sr >>= 3;
                xmc[38] = sr & 0x7;
                sr >>= 3;
                sr |= (uint16_t) *c++ << 4;
                Nc[3] = sr & 0x7F;
                sr >>= 7;
                bc[3] = sr & 0x3;
                sr >>= 2;
                Mc[3] = sr & 0x3;
                sr >>= 2;
                sr |= (uint16_t) *c++ << 1;
                xmaxc[3] = sr & 0x3F;
                sr >>= 6;
                xmc[39] = sr & 0x7;
                sr >>= 3;
                sr = *c++;
                xmc[40] = sr & 0x7;
                sr >>= 3;
                xmc[41] = sr & 0x7;
                sr >>= 3;
                sr |= (uint16_t) *c++ << 2;                              /* 30 */
                xmc[42] = sr & 0x7;
                sr >>= 3;
                xmc[43] = sr & 0x7;
                sr >>= 3;
                xmc[44] = sr & 0x7;
                sr >>= 3;
                sr |= (uint16_t) *c++ << 1;
                xmc[45] = sr & 0x7;
                sr >>= 3;
                xmc[46] = sr & 0x7;
                sr >>= 3;
                xmc[47] = sr & 0x7;
                sr >>= 3;
                sr = *c++;
                xmc[48] = sr & 0x7;
                sr >>= 3;
                xmc[49] = sr & 0x7;
                sr >>= 3;
                sr |= (uint16_t) *c++ << 2;
                xmc[50] = sr & 0x7;
                sr >>= 3;
                xmc[51] = sr & 0x7;
                sr >>= 3;
                s->frame_chain = sr & 0xF;
            }
            else
            {
                sr = s->frame_chain;
                sr |= (uint16_t) *c++ << 4;                              /* 1 */
                LARc[0] = sr & 0x3F;
                sr >>= 6;
                LARc[1] = sr & 0x3F;
                sr >>= 6;
                sr = *c++;
                LARc[2] = sr & 0x1F;
                sr >>= 5;
                sr |= (uint16_t) *c++ << 3;
                LARc[3] = sr & 0x1F;
                sr >>= 5;
                LARc[4] = sr & 0xF;
                sr >>= 4;
                sr |= (uint16_t) *c++ << 2;
                LARc[5] = sr & 0xF;
                sr >>= 4;
                LARc[6] = sr & 0x7;
                sr >>= 3;
                LARc[7] = sr & 0x7;
                sr >>= 3;
                sr = *c++;                                               /* 5 */
                Nc[0] = sr & 0x7F;
                sr >>= 7;
                sr |= (uint16_t) *c++ << 1;
                bc[0] = sr & 0x3;
                sr >>= 2;
                Mc[0] = sr & 0x3;
                sr >>= 2;
                sr |= (uint16_t) *c++ << 5;
                xmaxc[0] = sr & 0x3F;
                sr >>= 6;
                xmc[0] = sr & 0x7;
                sr >>= 3;
                xmc[1] = sr & 0x7;
                sr >>= 3;
                sr |= (uint16_t) *c++ << 1;
                xmc[2] = sr & 0x7;
                sr >>= 3;
                xmc[3] = sr & 0x7;
                sr >>= 3;
                xmc[4] = sr & 0x7;
                sr >>= 3;
                sr = *c++;
                xmc[5] = sr & 0x7;
                sr >>= 3;
                xmc[6] = sr & 0x7;
                sr >>= 3;
                sr |= (uint16_t) *c++ << 2;                             /* 10 */
                xmc[7] = sr & 0x7;
                sr >>= 3;
                xmc[8] = sr & 0x7;
                sr >>= 3;
                xmc[9] = sr & 0x7;
                sr >>= 3;
                sr |= (uint16_t) *c++ << 1;
                xmc[10] = sr & 0x7;
                sr >>= 3;
                xmc[11] = sr & 0x7;
                sr >>= 3;
                xmc[12] = sr & 0x7;
                sr >>= 3;
                sr = *c++;
                Nc[1] = sr & 0x7F;
                sr >>= 7;
                sr |= (uint16_t) *c++ << 1;
                bc[1] = sr & 0x3;
                sr >>= 2;
                Mc[1] = sr & 0x3;
                sr >>= 2;
                sr |= (uint16_t) *c++ << 5;
                xmaxc[1] = sr & 0x3F;
                sr >>= 6;
                xmc[13] = sr & 0x7;
                sr >>= 3;
                xmc[14] = sr & 0x7;
                sr >>= 3;
                sr |= (uint16_t) *c++ << 1;                             /* 15 */
                xmc[15] = sr & 0x7;
                sr >>= 3;
                xmc[16] = sr & 0x7;
                sr >>= 3;
                xmc[17] = sr & 0x7;
                sr >>= 3;
                sr = *c++;
                xmc[18] = sr & 0x7;
                sr >>= 3;
                xmc[19] = sr & 0x7;
                sr >>= 3;
                sr |= (uint16_t) *c++ << 2;
                xmc[20] = sr & 0x7;
                sr >>= 3;
                xmc[21] = sr & 0x7;
                sr >>= 3;
                xmc[22] = sr & 0x7;
                sr >>= 3;
                sr |= (uint16_t) *c++ << 1;
                xmc[23] = sr & 0x7;
                sr >>= 3;
                xmc[24] = sr & 0x7;
                sr >>= 3;
                xmc[25] = sr & 0x7;
                sr >>= 3;
                sr = *c++;
                Nc[2] = sr & 0x7F;
                sr >>= 7;
                sr |= (uint16_t) *c++ << 1;                             /* 20 */
                bc[2] = sr & 0x3;
                sr >>= 2;
                Mc[2] = sr & 0x3;
                sr >>= 2;
                sr |= (uint16_t) *c++ << 5;
                xmaxc[2] = sr & 0x3F;
                sr >>= 6;
                xmc[26] = sr & 0x7;
                sr >>= 3;
                xmc[27] = sr & 0x7;
                sr >>= 3;
                sr |= (uint16_t) *c++ << 1;
                xmc[28] = sr & 0x7;
                sr >>= 3;
                xmc[29] = sr & 0x7;
                sr >>= 3;
                xmc[30] = sr & 0x7;
                sr >>= 3;
                sr = *c++;
                xmc[31] = sr & 0x7;
                sr >>= 3;
                xmc[32] = sr & 0x7;
                sr >>= 3;
                sr |= (uint16_t) *c++ << 2;
                xmc[33] = sr & 0x7;
                sr >>= 3;
                xmc[34] = sr & 0x7;
                sr >>= 3;
                xmc[35] = sr & 0x7;
                sr >>= 3;
                sr |= (uint16_t) *c++ << 1;                             /* 25 */
                xmc[36] = sr & 0x7;
                sr >>= 3;
                xmc[37] = sr & 0x7;
                sr >>= 3;
                xmc[38] = sr & 0x7;
                sr >>= 3;
                sr = *c++;
                Nc[3] = sr & 0x7f;
                sr >>= 7;
                sr |= (uint16_t) *c++ << 1;
                bc[3] = sr & 0x3;
                sr >>= 2;
                Mc[3] = sr & 0x3;
                sr >>= 2;
                sr |= (uint16_t) *c++ << 5;
                xmaxc[3] = sr & 0x3F;
                sr >>= 6;
                xmc[39] = sr & 0x7;
                sr >>= 3;
                xmc[40] = sr & 0x7;
                sr >>= 3;
                sr |= (uint16_t) *c++ << 1;
                xmc[41] = sr & 0x7;
                sr >>= 3;
                xmc[42] = sr & 0x7;
                sr >>= 3;
                xmc[43] = sr & 0x7;
                sr >>= 3;
                sr = *c++;                                              /* 30 */
                xmc[44] = sr & 0x7;
                sr >>= 3;
                xmc[45] = sr & 0x7;
                sr >>= 3;
                sr |= (uint16_t) *c++ << 2;
                xmc[46] = sr & 0x7;
                sr >>= 3;
                xmc[47] = sr & 0x7;
                sr >>= 3;
                xmc[48] = sr & 0x7;
                sr >>= 3;
                sr |= (uint16_t) *c++ << 1;
                xmc[49] = sr & 0x7;
                sr >>= 3;
                xmc[50] = sr & 0x7;
                sr >>= 3;
                xmc[51] = sr & 0x7;
                sr >>= 3;
            }
            /*endif*/
            break;
        case GSM0610_PACKING_VOIP:
            if (((c[0] >> 4) & 0x0F) != GSM0610_MAGIC)
                break;
            LARc[0] = ((c[0] & 0xF) << 2) | ((c[1] >> 6) & 0x3);
            LARc[1] = c[1] & 0x3F;
            LARc[2] = (c[2] >> 3) & 0x1F;
            LARc[3] = ((c[2] & 0x7) << 2) | ((c[3] >> 6) & 0x3);
            LARc[4] = (c[3] >> 2) & 0xF;
            LARc[5] = ((c[3] & 0x3) << 2) | ((c[4] >> 6) & 0x3);
            LARc[6] = (c[4] >> 3) & 0x7;
            LARc[7] = c[4] & 0x7;
            Nc[0] = (c[5] >> 1) & 0x7F;
            bc[0] = ((c[5] & 0x1) << 1) | ((c[6] >> 7) & 0x1);
            Mc[0] = (c[6] >> 5) & 0x3;
            xmaxc[0] = ((c[6] & 0x1F) << 1) | ((c[7] >> 7) & 0x1);
            xmc[0] = (c[7] >> 4) & 0x7;
            xmc[1] = (c[7] >> 1) & 0x7;
            xmc[2] = ((c[7] & 0x1) << 2) | (((c[8] >> 6) & 0x3) | ((c[8] >> 3) & 0x7));
            xmc[4] = c[8] & 0x7;
            xmc[5] = (c[9] >> 5) & 0x7;
            xmc[6] = (c[9] >> 2) & 0x7;
            xmc[7] = ((c[9] & 0x3) << 1) | ((c[10] >> 7) & 0x1);
            xmc[8] = (c[10] >> 4) & 0x7;
            xmc[9] = (c[10] >> 1) & 0x7;
            xmc[10] = ((c[10] & 0x1) << 2) | ((c[11] >> 6) & 0x3);
            xmc[11] = (c[11] >> 3) & 0x7;
            xmc[12] = c[11] & 0x7;
            Nc[1] = (c[12] >> 1) & 0x7F;
            bc[1] = ((c[12] & 0x1) << 1) | ((c[13] >> 7) & 0x1);
            Mc[1] = (c[13] >> 5) & 0x3;
            xmaxc[1] = ((c[13] & 0x1F) << 1) | ((c[14] >> 7) & 0x1);
            xmc[13] = (c[14] >> 4) & 0x7;
            xmc[14] = (c[14] >> 1) & 0x7;
            xmc[15] = ((c[14] & 0x1) << 2) | ((c[15] >> 6) & 0x3);
            xmc[16] = (c[15] >> 3) & 0x7;
            xmc[17] = c[15] & 0x7;
            xmc[18] = (c[16] >> 5) & 0x7;
            xmc[19] = (c[16] >> 2) & 0x7;
            xmc[20] = ((c[16] & 0x3) << 1) | ((c[17] >> 7) & 0x1);
            xmc[21] = (c[17] >> 4) & 0x7;
            xmc[22] = (c[17] >> 1) & 0x7;
            xmc[23] = ((c[17] & 0x1) << 2) | ((c[18] >> 6) & 0x3);
            xmc[24] = (c[18] >> 3) & 0x7;
            xmc[25] = c[18] & 0x7;
            Nc[2] = (c[19] >> 1) & 0x7F;
            bc[2] = ((c[19] & 0x1) << 1) | ((c[20] >> 7) & 0x1);
            Mc[2] = (c[20] >> 5) & 0x3;
            xmaxc[2] = ((c[20] & 0x1F) << 1) | ((c[21] >> 7) & 0x1);
            xmc[26] = (c[21] >> 4) & 0x7;
            xmc[27] = (c[21] >> 1) & 0x7;
            xmc[28] = ((c[21] & 0x1) << 2) | ((c[22] >> 6) & 0x3);
            xmc[29] = (c[22] >> 3) & 0x7;
            xmc[30] = c[22] & 0x7;
            xmc[31] = (c[23] >> 5) & 0x7;
            xmc[32] = (c[23] >> 2) & 0x7;
            xmc[33] = ((c[23] & 0x3) << 1) | ((c[24] >> 7) & 0x1);
            xmc[34] = (c[24] >> 4) & 0x7;
            xmc[35] = (c[24] >> 1) & 0x7;
            xmc[36] = ((c[24] & 0x1) << 2) | ((c[25] >> 6) & 0x3);
            xmc[37] = (c[25] >> 3) & 0x7;
            xmc[38] = c[25] & 0x7;
            Nc[3] = (c[26] >> 1) & 0x7F;
            bc[3] = ((c[26] & 0x1) << 1) | ((c[27] >> 7) & 0x1);
            Mc[3] = (c[27] >> 5) & 0x3;
            xmaxc[3] = ((c[27] & 0x1F) << 1) | ((c[28] >> 7) & 0x1);
            xmc[39] = (c[28] >> 4) & 0x7;
            xmc[40] = (c[28] >> 1) & 0x7;
            xmc[41] = ((c[28] & 0x1) << 2) | ((c[29] >> 6) & 0x3);
            xmc[42] = (c[29] >> 3) & 0x7;
            xmc[43] = c[29] & 0x7;
            xmc[44] = (c[30] >> 5) & 0x7;
            xmc[45] = (c[30] >> 2) & 0x7;
            xmc[46] = ((c[30] & 0x3) << 1) | ((c[31] >> 7) & 0x1);
            xmc[47] = (c[31] >> 4) & 0x7;
            xmc[48] = (c[31] >> 1) & 0x7;
            xmc[49] = ((c[31] & 0x1) << 2) | ((c[32] >> 6) & 0x3);
            xmc[50] = (c[32] >> 3) & 0x7;
            xmc[51] = c[32] & 0x7;
            break;
        }
        /*endswitch*/
        if (s->log_gsm_block)
        {
            fprintf(stderr,
                    "LARc:\t%.2d %.2d %.2d %.2d %.2d %.2d %.2d %.2d\n",
                    LARc[0],
                    LARc[1],
                    LARc[2],
                    LARc[3],
                    LARc[4],
                    LARc[5],
                    LARc[6],
                    LARc[7]);
    
            for (i = 0;  i < 4;  i++)
            {
                fprintf(stderr,
                        "#%d:         Nc %4.4d    bc %d    Mc %d    xmaxc %d\n",
                        i + 1,
                        Nc[i],
                        bc[i],
                        Mc[i],
                        xmaxc[i]);
                fprintf(stderr,
                        "\t%.2d %.2d %.2d %.2d %.2d %.2d %.2d %.2d %.2d %.2d %.2d %.2d %.2d\n",
                        xmc[i*13 + 0],
                        xmc[i*13 + 1],
                        xmc[i*13 + 2],
                        xmc[i*13 + 3],
                        xmc[i*13 + 4],
                        xmc[i*13 + 5],
                        xmc[i*13 + 6],
                        xmc[i*13 + 7],
                        xmc[i*13 + 8],
                        xmc[i*13 + 9],
                        xmc[i*13 + 10],
                        xmc[i*13 + 11],
                        xmc[i*13 + 12]);
            }
            /*endfor*/
        }
        /*endif*/
    
        decode_a_frame(s, LARc, Nc, bc, Mc, xmaxc, xmc, amp);
        amp += GSM0610_FRAME_LEN;
    }
    /*endwhile*/
    return c;
}
/*- End of function --------------------------------------------------------*/
/*- End of file ------------------------------------------------------------*/
