/*
 * SpanDSP - a series of DSP components for telephony
 *
 * gsm0610_encode.c - GSM 06.10 full rate speech codec.
 *
 * Written by Steve Underwood <steveu@coppice.org>
 *
 * Copyright (C) 2006 Steve Underwood
 *
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * This code is based on the widely used GSM 06.10 code available from
 * http://kbs.cs.tu-berlin.de/~jutta/toast.html
 *
 * $Id: gsm0610_encode.c,v 1.4 2006/06/13 13:29:46 steveu Exp $
 */

/*! \file */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <assert.h>
#include <inttypes.h>
#include <tgmath.h>
#include <stdlib.h>
#include <memory.h>
#include <stdio.h>

#include "spandsp/telephony.h"
#include "spandsp/dc_restore.h"
#include "spandsp/gsm0610.h"

#include "gsm0610_local.h"

/* 4.2 FIXED POINT IMPLEMENTATION OF THE RPE-LTP CODER */

/* The RPE-LTD coder works on a frame by frame basis.  The length of
   the frame is equal to 160 samples.  Some computations are done
   once per frame to produce at the output of the coder the
   LARc[1..8] parameters which are the coded LAR coefficients and 
   also to realize the inverse filtering operation for the entire
   frame (160 samples of signal d[0..159]).  These parts produce at
   the output of the coder:

   Procedure 4.2.11 to 4.2.18 are to be executed four times per
   frame.  That means once for each sub-segment RPE-LTP analysis of
   40 samples.  These parts produce at the output of the coder.
*/
static void encode(gsm0610_state_t *s,
                    const int16_t amp[GSM0610_FRAME_LEN],       // samples                    IN
                    int16_t LARc[8],      // LAR coefficients           OUT
                    int16_t Nc[4],        // LTP lag                    OUT
                    int16_t bc[4],        // coded LTP gain             OUT
                    int16_t Mc[4],        // RPE grid selection         OUT
                    int16_t xmaxc[4],     // Coded maximum amplitude    OUT
                    int16_t *xMc)         // [13*4] normalized RPE samples     OUT
{
    int k;
    int16_t *dp;
    int16_t *dpp;
    int16_t so[GSM0610_FRAME_LEN];
    int16_t *Ncx;
    int16_t *bcx;
    int16_t *Mcx;
    int16_t *xmaxcx;
    int i;

    Ncx = Nc;
    bcx = bc;
    Mcx = Mc;
    xmaxcx = xmaxc;

    dp = s->dp0 + 120;
    dpp = dp;
    gsm0610_preprocess(s, amp, so);
    gsm0610_lpc_analysis(s, so, LARc);
    gsm0610_short_term_analysis_filter(s, LARc, so);

    for (k = 0;  k < 4;  k++, xMc += 13)
    {
        gsm0610_long_term_predictor(s,
                                    so + k*40,
                                    dp,
                                    s->e + 5,
                                    dpp,
                                    Ncx++,
                                    bcx++);

        gsm0610_rpe_encoding(s, s->e + 5, xmaxcx++, Mcx++, xMc);

        for (i = 0;  i < 40;  i++)
            dp[i] = gsm_add(s->e[5 + i], dpp[i]);
        /*endfor*/
        dp += 40;
        dpp += 40;
    }
    /*endfor*/
    memcpy ((char *) s->dp0,
            (char *) (s->dp0 + GSM0610_FRAME_LEN),
            120*sizeof(*s->dp0));
}
/*- End of function --------------------------------------------------------*/

gsm0610_state_t *gsm0610_init(gsm0610_state_t *s, int packing)
{
    if (s == NULL)
    {
        s = (gsm0610_state_t *) malloc(sizeof (gsm0610_state_t));
        if (s == NULL)
            return  NULL;
        /*endif*/
    }
    /*endif*/
    memset((char *) s, '\0', sizeof (gsm0610_state_t));
    s->nrp = 40;
    s->packing = packing;
    return s;
}
/*- End of function --------------------------------------------------------*/

int gsm0610_release(gsm0610_state_t *s)
{
    if (s)
        free(s);
    /*endif*/
    return 0;
}
/*- End of function --------------------------------------------------------*/

uint8_t *gsm0610_encode(gsm0610_state_t *s, uint8_t c[], const int16_t amp[], int quant)
{
    int16_t LARc[8];
    int16_t Nc[4];
    int16_t Mc[4];
    int16_t bc[4];
    int16_t xmaxc[4];
    int16_t xmc[13*4];
    uint16_t sr;
    int i;

    if (s->packing == 1)
        quant <<= 1;
    while (quant--)
    {
        encode(s, amp, LARc, Nc, bc, Mc, xmaxc, xmc);
        if (s->log_gsm_block)
        {
            fprintf(stderr,
                    "LARc:\t%.2d %.2d %.2d %.2d %.2d %.2d %.2d %.2d\n",
                    LARc[0],
                    LARc[1],
                    LARc[2],
                    LARc[3],
                    LARc[4],
                    LARc[5],
                    LARc[6],
                    LARc[7]);

            for (i = 0;  i < 4;  i++)
            {
                fprintf(stderr,
                        "#%d:         Nc %4.4d    bc %d    Mc %d    xmaxc %d\n",
                        i + 1,
                        Nc[i],
                        bc[i],
                        Mc[i],
                        xmaxc[i]);
                fprintf(stderr,
                        "\t%.2d %.2d %.2d %.2d %.2d %.2d %.2d %.2d %.2d %.2d %.2d %.2d %.2d\n",
                        xmc[i*13 + 0],
                        xmc[i*13 + 1],
                        xmc[i*13 + 2],
                        xmc[i*13 + 3],
                        xmc[i*13 + 4],
                        xmc[i*13 + 5],
                        xmc[i*13 + 6],
                        xmc[i*13 + 7],
                        xmc[i*13 + 8],
                        xmc[i*13 + 9],
                        xmc[i*13 + 10],
                        xmc[i*13 + 11],
                        xmc[i*13 + 12]);
            }
            /*endfor*/
        }
        /*endif*/
    
        /*  variable    size

            LARc[0]     6
            LARc[1]     6
            LARc[2]     5
            LARc[3]     5
            LARc[4]     4
            LARc[5]     4
            LARc[6]     3
            LARc[7]     3
    
            Nc[0]       7
            bc[0]       2
            Mc[0]       2
            xmaxc[0]    6
            xmc[0]      3
            xmc[1]      3
            xmc[2]      3
            xmc[3]      3
            xmc[4]      3
            xmc[5]      3
            xmc[6]      3
            xmc[7]      3
            xmc[8]      3
            xmc[9]      3
            xmc[10]     3
            xmc[11]     3
            xmc[12]     3

            Nc[1]       7
            bc[1]       2
            Mc[1]       2
            xmaxc[1]    6
            xmc[13]     3
            xmc[14]     3
            xmc[15]     3
            xmc[16]     3
            xmc[17]     3
            xmc[18]     3
            xmc[19]     3
            xmc[20]     3
            xmc[21]     3
            xmc[22]     3
            xmc[23]     3
            xmc[24]     3
            xmc[25]     3

            Nc[2]       7
            bc[2]       2
            Mc[2]       2
            xmaxc[2]    6
            xmc[26]     3
            xmc[27]     3
            xmc[28]     3
            xmc[29]     3
            xmc[30]     3
            xmc[31]     3
            xmc[32]     3
            xmc[33]     3
            xmc[34]     3
            xmc[35]     3
            xmc[36]     3
            xmc[37]     3
            xmc[38]     3

            Nc[3]       7
            bc[3]       2
            Mc[3]       2
            xmaxc[3]    6
            xmc[39]     3
            xmc[40]     3
            xmc[41]     3
            xmc[42]     3
            xmc[43]     3
            xmc[44]     3
            xmc[45]     3
            xmc[46]     3
            xmc[47]     3
            xmc[48]     3
            xmc[49]     3
            xmc[50]     3
            xmc[51]     3
        */

        switch (s->packing)
        {
        case GSM0610_PACKING_NONE:
            /* No packing at all */
            c[0] = LARc[0];
            c[1] = LARc[1];
            c[2] = LARc[2];
            c[3] = LARc[3];
            c[4] = LARc[4];
            c[5] = LARc[5];
            c[6] = LARc[6];
            c[7] = LARc[7];
            c[8] = Nc[0];
            c[9] = bc[0];
            c[10] = Mc[0];
            c[11] = xmaxc[0];
            c[12] = xmc[0];
            c[13] = xmc[1];
            c[14] = xmc[2];
            c[15] = xmc[3];
            c[16] = xmc[4];
            c[17] = xmc[5];
            c[18] = xmc[6];
            c[19] = xmc[7];
            c[20] = xmc[8];
            c[21] = xmc[9];
            c[22] = xmc[10];
            c[23] = xmc[11];
            c[24] = xmc[12];
            c[25] = Nc[1];
            c[26] = bc[1];
            c[27] = Mc[1];
            c[28] = xmaxc[1];
            c[29] = xmc[13];
            c[30] = xmc[14];
            c[31] = xmc[15];
            c[32] = xmc[16];
            c[33] = xmc[17];
            c[34] = xmc[18];
            c[35] = xmc[19];
            c[36] = xmc[20];
            c[37] = xmc[21];
            c[38] = xmc[22];
            c[39] = xmc[23];
            c[40] = xmc[24];
            c[41] = xmc[25];
            c[42] = Nc[2];
            c[43] = bc[2];
            c[44] = Mc[2];
            c[45] = xmaxc[2];
            c[46] = xmc[26];
            c[47] = xmc[27];
            c[48] = xmc[28];
            c[49] = xmc[29];
            c[50] = xmc[30];
            c[51] = xmc[31];
            c[52] = xmc[32];
            c[53] = xmc[33];
            c[54] = xmc[34];
            c[55] = xmc[35];
            c[56] = xmc[36];
            c[57] = xmc[37];
            c[58] = xmc[38];
            c[59] = Nc[3];
            c[60] = bc[3];
            c[61] = Mc[3];
            c[62] = xmaxc[3];
            c[63] = xmc[39];
            c[64] = xmc[40];
            c[65] = xmc[41];
            c[66] = xmc[42];
            c[67] = xmc[43];
            c[68] = xmc[44];
            c[69] = xmc[45];
            c[70] = xmc[46];
            c[71] = xmc[47];
            c[72] = xmc[48];
            c[73] = xmc[49];
            c[74] = xmc[50];
            c[75] = xmc[51];
            c += 76;
            break;
        case GSM0610_PACKING_WAV49:
            s->frame_index = !s->frame_index;
            if (s->frame_index)
            {
                sr = 0;
                sr = sr >> 6 | LARc[0] << 10;
                sr = sr >> 6 | LARc[1] << 10;
                *c++ = sr >> 4;
                sr = sr >> 5 | LARc[2] << 11;
                *c++ = sr >> 7;
                sr = sr >> 5 | LARc[3] << 11;
                sr = sr >> 4 | LARc[4] << 12;
                *c++ = sr >> 6;
                sr = sr >> 4 | LARc[5] << 12;
                sr = sr >> 3 | LARc[6] << 13;
                *c++ = sr >> 7;
                sr = sr >> 3 | LARc[7] << 13;
                sr = sr >> 7 | Nc[0] << 9;
                *c++ = sr >> 5;
                sr = sr >> 2 | bc[0] << 14;
                sr = sr >> 2 | Mc[0] << 14;
                sr = sr >> 6 | xmaxc[0] << 10;
                *c++ = sr >> 3;
                sr = sr >> 3 | xmc[0] << 13;
                *c++ = sr >> 8;
                sr = sr >> 3 | xmc[1] << 13;
                sr = sr >> 3 | xmc[2] << 13;
                sr = sr >> 3 | xmc[3] << 13;
                *c++ = sr >> 7;
                sr = sr >> 3 | xmc[4] << 13;
                sr = sr >> 3 | xmc[5] << 13;
                sr = sr >> 3 | xmc[6] << 13;
                *c++ = sr >> 6;
                sr = sr >> 3 | xmc[7] << 13;
                sr = sr >> 3 | xmc[8] << 13;
                *c++ = sr >> 8;
                sr = sr >> 3 | xmc[9] << 13;
                sr = sr >> 3 | xmc[10] << 13;
                sr = sr >> 3 | xmc[11] << 13;
                *c++ = sr >> 7;
                sr = sr >> 3 | xmc[12] << 13;
                sr = sr >> 7 | Nc[1] << 9;
                *c++ = sr >> 5;
                sr = sr >> 2 | bc[1] << 14;
                sr = sr >> 2 | Mc[1] << 14;
                sr = sr >> 6 | xmaxc[1] << 10;
                *c++ = sr >> 3;
                sr = sr >> 3 | xmc[13] << 13;
                *c++ = sr >> 8;
                sr = sr >> 3 | xmc[14] << 13;
                sr = sr >> 3 | xmc[15] << 13;
                sr = sr >> 3 | xmc[16] << 13;
                *c++ = sr >> 7;
                sr = sr >> 3 | xmc[17] << 13;
                sr = sr >> 3 | xmc[18] << 13;
                sr = sr >> 3 | xmc[19] << 13;
                *c++ = sr >> 6;
                sr = sr >> 3 | xmc[20] << 13;
                sr = sr >> 3 | xmc[21] << 13;
                *c++ = sr >> 8;
                sr = sr >> 3 | xmc[22] << 13;
                sr = sr >> 3 | xmc[23] << 13;
                sr = sr >> 3 | xmc[24] << 13;
                *c++ = sr >> 7;
                sr = sr >> 3 | xmc[25] << 13;
                sr = sr >> 7 | Nc[2] << 9;
                *c++ = sr >> 5;
                sr = sr >> 2 | bc[2] << 14;
                sr = sr >> 2 | Mc[2] << 14;
                sr = sr >> 6 | xmaxc[2] << 10;
                *c++ = sr >> 3;
                sr = sr >> 3 | xmc[26] << 13;
                *c++ = sr >> 8;
                sr = sr >> 3 | xmc[27] << 13;
                sr = sr >> 3 | xmc[28] << 13;
                sr = sr >> 3 | xmc[29] << 13;
                *c++ = sr >> 7;
                sr = sr >> 3 | xmc[30] << 13;
                sr = sr >> 3 | xmc[31] << 13;
                sr = sr >> 3 | xmc[32] << 13;
                *c++ = sr >> 6;
                sr = sr >> 3 | xmc[33] << 13;
                sr = sr >> 3 | xmc[34] << 13;
                *c++ = sr >> 8;
                sr = sr >> 3 | xmc[35] << 13;
                sr = sr >> 3 | xmc[36] << 13;
                sr = sr >> 3 | xmc[37] << 13;
                *c++ = sr >> 7;
                sr = sr >> 3 | xmc[38] << 13;
                sr = sr >> 7 | Nc[3] << 9;
                *c++ = sr >> 5;
                sr = sr >> 2 | bc[3] << 14;
                sr = sr >> 2 | Mc[3] << 14;
                sr = sr >> 6 | xmaxc[3] << 10;
                *c++ = sr >> 3;
                sr = sr >> 3 | xmc[39] << 13;
                *c++ = sr >> 8;
                sr = sr >> 3 | xmc[40] << 13;
                sr = sr >> 3 | xmc[41] << 13;
                sr = sr >> 3 | xmc[42] << 13;
                *c++ = sr >> 7;
                sr = sr >> 3 | xmc[43] << 13;
                sr = sr >> 3 | xmc[44] << 13;
                sr = sr >> 3 | xmc[45] << 13;
                *c++ = sr >> 6;
                sr = sr >> 3 | xmc[46] << 13;
                sr = sr >> 3 | xmc[47] << 13;
                *c++ = sr >> 8;
                sr = sr >> 3 | xmc[48] << 13;
                sr = sr >> 3 | xmc[49] << 13;
                sr = sr >> 3 | xmc[50] << 13;
                *c++ = sr >> 7;
                sr = sr >> 3 | xmc[51] << 13;
                sr = sr >> 4;
                *c = sr >> 8;
                s->frame_chain = *c;
            }
            else
            {
                sr = 0;
                sr = sr >> 4 | s->frame_chain << 12;
                sr = sr >> 6 | LARc[0] << 10;
                *c++ = sr >> 6;
                sr = sr >> 6 | LARc[1] << 10;
                *c++ = sr >> 8;
                sr = sr >> 5 | LARc[2] << 11;
                sr = sr >> 5 | LARc[3] << 11;
                *c++ = sr >> 6;
                sr = sr >> 4 | LARc[4] << 12;
                sr = sr >> 4 | LARc[5] << 12;
                *c++ = sr >> 6;
                sr = sr >> 3 | LARc[6] << 13;
                sr = sr >> 3 | LARc[7] << 13;
                *c++ = sr >> 8;
                sr = sr >> 7 | Nc[0] << 9;
                sr = sr >> 2 | bc[0] << 14;
                *c++ = sr >> 7;
                sr = sr >> 2 | Mc[0] << 14;
                sr = sr >> 6 | xmaxc[0] << 10;
                *c++ = sr >> 7;
                sr = sr >> 3 | xmc[0] << 13;
                sr = sr >> 3 | xmc[1] << 13;
                sr = sr >> 3 | xmc[2] << 13;
                *c++ = sr >> 6;
                sr = sr >> 3 | xmc[3] << 13;
                sr = sr >> 3 | xmc[4] << 13;
                *c++ = sr >> 8;
                sr = sr >> 3 | xmc[5] << 13;
                sr = sr >> 3 | xmc[6] << 13;
                sr = sr >> 3 | xmc[7] << 13;
                *c++ = sr >> 7;
                sr = sr >> 3 | xmc[8] << 13;
                sr = sr >> 3 | xmc[9] << 13;
                sr = sr >> 3 | xmc[10] << 13;
                *c++ = sr >> 6;
                sr = sr >> 3 | xmc[11] << 13;
                sr = sr >> 3 | xmc[12] << 13;
                *c++ = sr >> 8;
                sr = sr >> 7 | Nc[1] << 9;
                sr = sr >> 2 | bc[1] << 14;
                *c++ = sr >> 7;
                sr = sr >> 2 | Mc[1] << 14;
                sr = sr >> 6 | xmaxc[1] << 10;
                *c++ = sr >> 7;
                sr = sr >> 3 | xmc[13] << 13;
                sr = sr >> 3 | xmc[14] << 13;
                sr = sr >> 3 | xmc[15] << 13;
                *c++ = sr >> 6;
                sr = sr >> 3 | xmc[16] << 13;
                sr = sr >> 3 | xmc[17] << 13;
                *c++ = sr >> 8;
                sr = sr >> 3 | xmc[18] << 13;
                sr = sr >> 3 | xmc[19] << 13;
                sr = sr >> 3 | xmc[20] << 13;
                *c++ = sr >> 7;
                sr = sr >> 3 | xmc[21] << 13;
                sr = sr >> 3 | xmc[22] << 13;
                sr = sr >> 3 | xmc[23] << 13;
                *c++ = sr >> 6;
                sr = sr >> 3 | xmc[24] << 13;
                sr = sr >> 3 | xmc[25] << 13;
                *c++ = sr >> 8;
                sr = sr >> 7 | Nc[2] << 9;
                sr = sr >> 2 | bc[2] << 14;
                *c++ = sr >> 7;
                sr = sr >> 2 | Mc[2] << 14;
                sr = sr >> 6 | xmaxc[2] << 10;
                *c++ = sr >> 7;
                sr = sr >> 3 | xmc[26] << 13;
                sr = sr >> 3 | xmc[27] << 13;
                sr = sr >> 3 | xmc[28] << 13;
                *c++ = sr >> 6;
                sr = sr >> 3 | xmc[29] << 13;
                sr = sr >> 3 | xmc[30] << 13;
                *c++ = sr >> 8;
                sr = sr >> 3 | xmc[31] << 13;
                sr = sr >> 3 | xmc[32] << 13;
                sr = sr >> 3 | xmc[33] << 13;
                *c++ = sr >> 7;
                sr = sr >> 3 | xmc[34] << 13;
                sr = sr >> 3 | xmc[35] << 13;
                sr = sr >> 3 | xmc[36] << 13;
                *c++ = sr >> 6;
                sr = sr >> 3 | xmc[37] << 13;
                sr = sr >> 3 | xmc[38] << 13;
                *c++ = sr >> 8;
                sr = sr >> 7 | Nc[3] << 9;
                sr = sr >> 2 | bc[3] << 14;
                *c++ = sr >> 7;
                sr = sr >> 2 | Mc[3] << 14;
                sr = sr >> 6 | xmaxc[3] << 10;
                *c++ = sr >> 7;
                sr = sr >> 3 | xmc[39] << 13;
                sr = sr >> 3 | xmc[40] << 13;
                sr = sr >> 3 | xmc[41] << 13;
                *c++ = sr >> 6;
                sr = sr >> 3 | xmc[42] << 13;
                sr = sr >> 3 | xmc[43] << 13;
                *c++ = sr >> 8;
                sr = sr >> 3 | xmc[44] << 13;
                sr = sr >> 3 | xmc[45] << 13;
                sr = sr >> 3 | xmc[46] << 13;
                *c++ = sr >> 7;
                sr = sr >> 3 | xmc[47] << 13;
                sr = sr >> 3 | xmc[48] << 13;
                sr = sr >> 3 | xmc[49] << 13;
                *c++ = sr >> 6;
                sr = sr >> 3 | xmc[50] << 13;
                sr = sr >> 3 | xmc[51] << 13;
                *c++ = sr >> 8;
            }
            /*endif*/
        case GSM0610_PACKING_VOIP:
            c[0] = ((GSM0610_MAGIC & 0xF) << 4) | ((LARc[0] >> 2) & 0xF);
            c[1] = ((LARc[0] & 0x3) << 6) | (LARc[1] & 0x3F);
            c[2] = ((LARc[2] & 0x1F) << 3) | ((LARc[3] >> 2) & 0x7);
            c[3] = ((LARc[3] & 0x3) << 6) | ((LARc[4] & 0xF) << 2) | ((LARc[5] >> 2) & 0x3);
            c[4] = ((LARc[5] & 0x3) << 6) | ((LARc[6] & 0x7) << 3) | (LARc[7] & 0x7);
            c[5] = ((Nc[0] & 0x7F) << 1) | ((bc[0] >> 1) & 0x1);
            c[6] = ((bc[0] & 0x1) << 7) | ((Mc[0] & 0x3) << 5) | ((xmaxc[0] >> 1) & 0x1F);
            c[7] = ((xmaxc[0] & 0x1) << 7) | ((xmc[0] & 0x7) << 4) | ((xmc[1] & 0x7) << 1) | ((xmc[2] >> 2) & 0x1);
            c[8] = ((xmc[2] & 0x3) << 6) | ((xmc[3] & 0x7) << 3) | (xmc[4] & 0x7);
            c[9] = ((xmc[5] & 0x7) << 5) | ((xmc[6] & 0x7) << 2) | ((xmc[7] >> 1) & 0x3);
            c[10] = ((xmc[7] & 0x1) << 7) | ((xmc[8] & 0x7) << 4) | ((xmc[9] & 0x7) << 1) | ((xmc[10] >> 2) & 0x1);
            c[11] = ((xmc[10] & 0x3) << 6) | ((xmc[11] & 0x7) << 3) | (xmc[12] & 0x7);
            c[12] = ((Nc[1] & 0x7F) << 1) | ((bc[1] >> 1) & 0x1);
            c[13] = ((bc[1] & 0x1) << 7) | ((Mc[1] & 0x3) << 5) | ((xmaxc[1] >> 1) & 0x1F);
            c[14] = ((xmaxc[1] & 0x1) << 7) | ((xmc[13] & 0x7) << 4) | ((xmc[14] & 0x7) << 1) | ((xmc[15] >> 2) & 0x1);
            c[15] = ((xmc[15] & 0x3) << 6) | ((xmc[16] & 0x7) << 3) | (xmc[17] & 0x7);
            c[16] = ((xmc[18] & 0x7) << 5) | ((xmc[19] & 0x7) << 2) | ((xmc[20] >> 1) & 0x3);
            c[17] = ((xmc[20] & 0x1) << 7) | ((xmc[21] & 0x7) << 4) | ((xmc[22] & 0x7) << 1) | ((xmc[23] >> 2) & 0x1);
            c[18] = ((xmc[23] & 0x3) << 6) | ((xmc[24] & 0x7) << 3) | (xmc[25] & 0x7);
            c[19] = ((Nc[2] & 0x7F) << 1) | ((bc[2] >> 1) & 0x1);
            c[20] = ((bc[2] & 0x1) << 7) | ((Mc[2] & 0x3) << 5) | ((xmaxc[2] >> 1) & 0x1F);
            c[21] = ((xmaxc[2] & 0x1) << 7) | ((xmc[26] & 0x7) << 4) | ((xmc[27] & 0x7) << 1) | ((xmc[28] >> 2) & 0x1);
            c[22] = ((xmc[28] & 0x3) << 6) | ((xmc[29] & 0x7) << 3) | (xmc[30] & 0x7);
            c[23] = ((xmc[31] & 0x7) << 5) | ((xmc[32] & 0x7) << 2) | ((xmc[33] >> 1) & 0x3);
            c[24] = ((xmc[33] & 0x1) << 7) | ((xmc[34] & 0x7) << 4) | ((xmc[35] & 0x7) << 1) | ((xmc[36] >> 2) & 0x1);
            c[25] = ((xmc[36] & 0x3) << 6) | ((xmc[37] & 0x7) << 3) | (xmc[38] & 0x7);
            c[26] = ((Nc[3] & 0x7F) << 1) | ((bc[3] >> 1) & 0x1);
            c[27] = ((bc[3] & 0x1) << 7) | ((Mc[3] & 0x3) << 5) | ((xmaxc[3] >> 1) & 0x1F);
            c[28] = ((xmaxc[3] & 0x1) << 7) | ((xmc[39] & 0x7) << 4) | ((xmc[40] & 0x7) << 1) | ((xmc[41] >> 2) & 0x1);
            c[29] = ((xmc[41] & 0x3) << 6) | ((xmc[42] & 0x7) << 3) | (xmc[43] & 0x7);
            c[30] = ((xmc[44] & 0x7) << 5) | ((xmc[45] & 0x7) << 2) | ((xmc[46] >> 1) & 0x3);
            c[31] = ((xmc[46] & 0x1) << 7) | ((xmc[47] & 0x7) << 4) | ((xmc[48] & 0x7) << 1) | ((xmc[49] >> 2) & 0x1);
            c[32] = ((xmc[49] & 0x3) << 6) | ((xmc[50] & 0x7) << 3) | (xmc[51] & 0x7);
            c += 33;
            break;
        }
        /*endswitch*/
        amp += GSM0610_FRAME_LEN;
    }
    /*endwhile*/
    return c;
}
/*- End of function --------------------------------------------------------*/
/*- End of file ------------------------------------------------------------*/
